import { HttpException, HttpStatus, Injectable, NotFoundException } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { generateStockId, isArraySame, urlSlug } from 'src/helper/common-helper';
import { Constants } from 'src/utils/constants';
import { ResponseUtils } from 'src/utils/response.utils';
import { CreateProductDto } from './dto/create-product.dto';
import { UpdateProductDto } from './dto/update-product.dto';
import { Product } from './entity/product.entity';
import { ProductsRepository } from './product.repository';
import { v4 as uuidv4 } from 'uuid';
import readXlsxFile from 'read-excel-file/node';
import { UploadedMulterFileI } from 'src/modules/SpacesModule/SpacesService';
import { DoSpacesService } from 'src/modules/SpacesModule/SpacesService/doSpacesService';
import { ProductColorVariantsRepository } from '../product-color-variants/product-color-variants.repository';
import { CreateProductColorVariantsDto } from '../product-color-variants/dto/create-product-color-variants.dto';
import { ProductImageGalleryRepository } from '../product-image-gallery/product-image-gallery.repository';
import { CreateProductImageGalleryDto } from '../product-image-gallery/dto/create-product-image-gallery.dto';
import { ColorsRepository } from '../color/color-repository';
import { ProductOffersRepository } from '../product-offers/product-offers.repository';
import { ProductOfferVariantsRepository } from '../product-offer-variants/product-offer-variants.repository';
import { CreateProductOfferVariantsDto } from '../product-offer-variants/dto/create-product-offer-variants.dto';
import { ProductStockImportRepository } from '../product-stock-import/product-stock-import.repository';
import { CreateImportProductStockDto } from './dto/create-import-product-stock.dto';
import { CreateColorDto } from '../color/dto/create-color.dto';
import { CreateProductStockImportDto } from '../product-stock-import/dto/create-product-stock-import.dto';
import { CreateConfirmStockInProductDto } from './dto/create-confirm-product-stock.dto';
import { ProductColorVariants } from '../product-color-variants/entities/product-color-variants.entity';
import { Color } from '../color/entity/color.entity';
import { CreateImportProductDto } from './dto/create-import-product.dto';
import { CreateProductDumpImportDto } from '../product-dump-import/dto/create-product-dump-import.dto';
import { ProductDumpImportRepository } from '../product-dump-import/product-dump-importrepository';
import { CreateConfirmImportProductDto } from './dto/create-confirm-import-product.dto';
import { HomePagePromotionSectionRepository } from 'src/modules/CRUD/home-page-promotion-section/home-page-promotion-section.repository';

@Injectable()
export class ProductService {
    constructor(
        @InjectRepository(ProductsRepository)
        private readonly repository: ProductsRepository,
        private readonly doSpaceService: DoSpacesService,
        @InjectRepository(ColorsRepository)
        private readonly colorsRepository: ColorsRepository,
        @InjectRepository(ProductColorVariantsRepository)
        private readonly productColorVariantsRepository: ProductColorVariantsRepository,
        @InjectRepository(ProductImageGalleryRepository)
        private readonly productImageGalleryRepository: ProductImageGalleryRepository,
        @InjectRepository(ProductOffersRepository)
        private readonly productOffersRepository: ProductOffersRepository,
        @InjectRepository(ProductOfferVariantsRepository)
        private readonly productOfferVariantsRepository: ProductOfferVariantsRepository,
        @InjectRepository(ProductStockImportRepository)
        private readonly productStockImportRepository: ProductStockImportRepository,
        @InjectRepository(ProductDumpImportRepository)
        private readonly productDumpImportRepository: ProductDumpImportRepository,
        @InjectRepository(HomePagePromotionSectionRepository)
        private readonly homePagePromotionSectionRepository: HomePagePromotionSectionRepository
    ) {
    }

    private cleanNameString(stringName: string) {
        return stringName.trim().replace(/[^a-zA-Z0-9 ]/g, '');
    }

    async create(
        dto: CreateProductDto,
        files: {
            productImages?: UploadedMulterFileI[],
            featuredImage?: UploadedMulterFileI
        }
    ): Promise<any> {
        // Processing featured image
        if (files && files.featuredImage) {
            const featuredImage: any = await this.doSpaceService.uploadFile(files.featuredImage[0], 'Product/');
            dto.featuredImage = featuredImage;
        } else dto.featuredImage = 'https://ganja-container.nyc3.cdn.digitaloceanspaces.com/default/default.png';

        let slug = urlSlug(dto.name);

        // Create a product entity
        const productEntity = {
            name: dto.name,
            price: dto.price,
            sku: dto.sku,
            stock: dto.stock,
            description: dto.description,
            urlSlug: slug,
            weight: dto.weight,
            featuredImage: dto.featuredImage,
            discountPrice: dto.discountPrice,
            cost: dto.cost,
            isDeactive: dto.isDeactive ? dto.isDeactive : 0,
            isReadyToShip: dto.isReadyToShip ? dto.isReadyToShip : 0,
            brandId: dto.brandId,
            brandName: dto.brandName
        }
        const product = await this.repository.createEntity(productEntity as CreateProductDto);

        // Processing product images
        if (files && files.productImages) {
            for (const productFile of files.productImages) {
                const productImageUrl: any = await this.doSpaceService.uploadFile(productFile, 'Product/');
                const productImage = {
                    productId: product.id,
                    productName: product.name,
                    image: productImageUrl
                }
                await this.productImageGalleryRepository.createEntity(productImage as CreateProductImageGalleryDto)
            }
        }

        // Processiong color 
        if (dto.color && dto.color.length > 0) {
            const colors = dto.color;
            for (const color of colors) {
                const colorData = await this.colorsRepository.findByColorName(color);
                if (colorData) {
                    if (Object.keys(colorData).length > 0) {
                        const colorVariant = {
                            productId: product.id,
                            productName: product.name,
                            colorId: colorData[0]['id'],
                            colorName: color,
                            colorCode: colorData[0]['colorCode']
                        }
                        await this.productColorVariantsRepository.createEntity(colorVariant as CreateProductColorVariantsDto);
                    }
                }
            }
        }

        // Processing offer
        if (dto.offers && dto.offers.length > 0) {
            const offers = dto.offers;
            for (const offer of offers) {
                const offerData = await this.productOffersRepository.findByTitle(offer);
                if (offerData) {
                    if (Object.keys(offerData).length > 0) {
                        const offerVariant = {
                            productId: product.id,
                            productName: product.name,
                            offerId: offerData[0]['id'],
                            offerTitle: offer
                        }
                        await this.productOfferVariantsRepository.createEntity(offerVariant as CreateProductOfferVariantsDto)
                    }
                }
            }
        }

        const colorVariants = await this.productColorVariantsRepository.findByFilterQuery({ productId: product.id });
        const offerVariants = await this.productOfferVariantsRepository.findByFilterQuery({ productId: product.id });
        const imageGallery = await this.productImageGalleryRepository.findByFilterQuery({ productId: product.id });

        const data = {
            ...product,
            colorVariants,
            offerVariants,
            imageGallery
        }

        return ResponseUtils.successResponseHandler(200, "Product created successfully!", "data", data);
    }

    async findAll(page: number, limit: number): Promise<any | Error> {
        const productData = await this.repository.findAllEntities(page, limit);

        return ResponseUtils.successResponseHandler(200, "Products fetched successfully", "data", productData);
    }

    async findOne(id: string): Promise<Product | Error> {
        const productSearch = await this.repository.findOneRawEntity(id);
        if (!productSearch) {
            throw new HttpException(
                Constants.NOT_FOUND,
                HttpStatus.NOT_FOUND
            );
        }
        const product = productSearch as Product;

        const colorVariants = await this.productColorVariantsRepository.findByFilterQuery({ productId: product.id });
        const offerVariants = await this.productOfferVariantsRepository.findByFilterQuery({ productId: product.id });
        const imageGallery = await this.productImageGalleryRepository.findByFilterQuery({ productId: product.id });

        const data = {
            ...product,
            colorVariants,
            offerVariants,
            imageGallery
        }
        return ResponseUtils.successResponseHandler(200, "Product data fetched Successfully!", "data", data);
    }

    async updateOneProduct(
        id: string,
        dto: UpdateProductDto,
        files: {
            productImages?: UploadedMulterFileI[],
            featuredImage?: UploadedMulterFileI
        }
    ): Promise<Product | Error> {
        if (files && files.featuredImage) {
            const featuredImage: any = await this.doSpaceService.uploadFile(files.featuredImage[0], 'Product/');
            dto.featuredImage = featuredImage;
        }

        const productObject = await this.repository.findOneRawEntity(id);
        if (!productObject) {
            throw new HttpException(
                Constants.NOT_FOUND,
                HttpStatus.NOT_FOUND
            );
        }

        let slug = urlSlug(dto.name);

        const productEntity = {
            name: dto.name,
            price: dto.price,
            sku: dto.sku,
            stock: dto.stock,
            description: dto.description,
            urlSlug: slug,
            weight: dto.weight,
            featuredImage: dto.featuredImage,
            discountPrice: dto.discountPrice,
            cost: dto.cost,
            isDeactive: dto.isDeactive ? dto.isDeactive : 0,
            isReadyToShip: dto.isReadyToShip ? dto.isReadyToShip : 0,
            brandId: dto.brandId,
            brandName: dto.brandName
        }
        const product = await this.repository.updateEntity(id, productEntity);

        // Processiong color 
        const currentColor = await this.productColorVariantsRepository.findByFilterQuery({ productId: id });
        if (dto.color && dto.color.length > 0) {
            const arrayCheck = await isArraySame(currentColor, dto.color);
            if (!arrayCheck) {
                const newColorEntity: any = [];
                for (const color of dto.color) {
                    const colorData = await this.colorsRepository.findByColorName(color);
                    if (colorData) {
                        if (Object.keys(colorData).length > 0) {
                            newColorEntity.push({
                                productId: id,
                                productName: productObject.name,
                                colorId: colorData[0]['id'],
                                colorName: color,
                                colorCode: colorData[0]['colorCode']
                            })
                        }
                    }
                }
                await this.productColorVariantsRepository.updateProductColorVariant(id, newColorEntity)
            }
        }

        // Processing offers
        const currentOffers = await this.productOfferVariantsRepository.findByFilterQuery({ productId: id });
        if (dto.offers && dto.offers.length > 0) {
            const arrayCheck = await isArraySame(currentOffers, dto.offers);
            if (!arrayCheck) {
                const newOffersEntity: any = [];
                for (const offer of dto.offers) {
                    const offerData = await this.productOffersRepository.findByTitle(offer);
                    if (offerData) {
                        if (Object.keys(offerData).length > 0) {
                            newOffersEntity.push({
                                productId: id,
                                productName: productObject.name,
                                offerId: offerData[0]['id'],
                                offerTitle: offer
                            })
                        }
                    }
                }
                await this.productOfferVariantsRepository.updateProductOfferVariant(id, newOffersEntity)
            }
        } else await this.productOfferVariantsRepository.deleteOfferVariant(id)

        if (files && files.productImages) {
            for (const productFile of files.productImages) {
                const productImageUrl: any = await this.doSpaceService.uploadFile(productFile, 'Product/');
                const productImage = {
                    productId: id,
                    productName: productObject.name,
                    image: productImageUrl
                }
                await this.productImageGalleryRepository.createEntity(productImage as CreateProductImageGalleryDto)
            }
        }


        if (dto.deletedImage && dto.deletedImage.length > 0) {
            await this.productImageGalleryRepository.deleteImage(dto.deletedImage);
        }

        const updatedProduct = product as Product;

        const colorVariants = await this.productColorVariantsRepository.findByFilterQuery({ productId: updatedProduct.id });
        const offerVariants = await this.productOfferVariantsRepository.findByFilterQuery({ productId: updatedProduct.id });
        const imageGallery = await this.productImageGalleryRepository.findByFilterQuery({ productId: updatedProduct.id });

        const data = {
            ...updatedProduct,
            colorVariants,
            offerVariants,
            imageGallery
        }

        return ResponseUtils.successResponseHandler(200, "Product updated successfully!", "product", data);
    }

    async remove(id: string): Promise<any | Error> {
        const data = await this.repository.deleteEntity(id);

        if (!data) {
            throw new HttpException(
                Constants.NOT_FOUND,
                HttpStatus.NOT_FOUND
            );
        }

        return ResponseUtils.buildDeletedData(data, 200, "Product data deleted successfully!");
    }

    async findOneBySlug(slug: string): Promise<Product | Error> {
        const findBySlug = await this.repository.findOneProductBySlug(slug);

        if (!findBySlug) {
            throw new HttpException(
                Constants.NOT_FOUND,
                HttpStatus.NOT_FOUND
            );
        }
        const product = findBySlug as Product;
        const colorVariants = await this.productColorVariantsRepository.findByFilterQuery({ productId: product.id });
        const offerVariants = await this.productOfferVariantsRepository.findByFilterQuery({ productId: product.id });
        const imageGallery = await this.productImageGalleryRepository.findByFilterQuery({ productId: product.id });

        const data = {
            ...product,
            colorVariants,
            offerVariants,
            imageGallery
        }
        return ResponseUtils.successResponseHandler(200, "Product data fetched Successfully!", "data", data);
    }

    async importProductStockService(dto: CreateImportProductStockDto, file: Express.Multer.File): Promise<any> {
        const response = await this.productStockImportRepository.findStockIdEntity();
        const StockID = response ? await generateStockId(response) : 'ganja-container';

        const excelDataRow = await readXlsxFile(file.path).then((rows) => {
            let excelDataRowDump = [];
            rows.forEach(async (element, index) => {
                if (index > 0) {
                    let productId: string = element[0].toString();
                    let productRawData = await {
                        'StockID': StockID,
                        'ProductId': productId,
                        'Name': element[1],
                        'Sku': element[2],
                        'Price': element[3],
                        'CurrentStock': 0,
                        'NewStock': element[4],
                        'NewTotalStock': 0,
                        'QuantityReorder': element[5],
                        'Weight': element[6],
                        'Color': element[7],
                        'ColorID': []
                    };
                    excelDataRowDump.push(productRawData);
                }
            });
            return excelDataRowDump;
        })

        for (let index = 0; index < excelDataRow.length; index++) {
            const rowsData = excelDataRow[index];

            // Colors Section
            let colorsArray = [];
            if (JSON.stringify(rowsData.Color).length > 0) {
                let colorsData = JSON.stringify(rowsData.Color).trim();
                colorsArray = colorsData.split(',');
            }

            for (let j = 0; j < colorsArray.length; j++) {
                const colorsName = colorsArray[j];

                let colorsRawData = await this.colorsRepository.findByFilterQuery({
                    colorName: this.cleanNameString(colorsName).trim()
                });

                if (!colorsRawData) {
                    let colorDto = {
                        colorName: this.cleanNameString(colorsName).trim(),
                        colorCode: '#0000'
                    };
                    let newColorData = await this.colorsRepository.createEntity(colorDto as CreateColorDto);
                    excelDataRow[index].ColorID.push(newColorData);
                } else {
                    if (Object.keys(colorsRawData).length == 0) {
                        let colorDto = {
                            colorName: this.cleanNameString(colorsName).trim(),
                            colorCode: '#0000'
                        };
                        let newColorData = await this.colorsRepository.createEntity(colorDto as CreateColorDto);
                        excelDataRow[index].ColorID.push(newColorData);
                    } else {
                        excelDataRow[index].ColorID.push(colorsRawData[0]);
                    }
                }
            }
        }

        for (let k = 0; k < excelDataRow.length; k++) {
            const element = excelDataRow[k];
            const id = element.ProductId;
            let productInfo = await this.repository.findOne(id);
            if (typeof productInfo === 'undefined') {

            } else {
                let currentQuantity = productInfo["stock"];
                let newQuantity = element.NewStock;
                let totalQuantity = currentQuantity + newQuantity;
                excelDataRow[k]['CurrentStock'] = currentQuantity;
                excelDataRow[k]['NewTotalStock'] = totalQuantity;
            }
        }

        excelDataRow.forEach(async row => {
            let dataRow = {
                'StockID': row.StockID,
                'ProductId': row.ProductId,
                'Name': row.Name,
                'Price': row.Price,
                'CurrentStock': row.CurrentStock,
                'NewStock': row.NewStock,
                'NewTotalStock': row.NewTotalStock,
                'QuantityReorder': row.QuantityReorder,
                'Sku': row.Sku,
                'Weight': row.Weight,
                'Color': row.Color
            };
            const data = await this.productStockImportRepository.createEntity(dataRow as CreateProductStockImportDto);
        });

        let returnData = {
            stockID: StockID,
            stock: excelDataRow
        }

        return ResponseUtils.successResponseHandler(200, "Import product stock successfully created!", "product_stock", returnData);
    }

    async confirmStockInProductService(dto: CreateConfirmStockInProductDto): Promise<any> {
        if (dto.stockID.length > 0) {
            let stockData = await this.productStockImportRepository.findByFilterQuery({
                StockID: dto.stockID
            });

            if (!stockData) {
                return ResponseUtils.errorResponseHandler(404, "Stock data not available!");
            }

            for (let index = 0; index < Object.keys(stockData).length; index++) {
                const row = stockData[index];
                let dataObject = {
                    'name': row.Name,
                    'price': row.Price,
                    'stock': row.NewTotalStock,
                    'weight': row.Weight,
                    'sku': row.Sku,
                    'color': row.Color
                }

                // Colors Section
                let colorsArray = [];
                if (JSON.stringify(row.Color).length > 0) {
                    let colorsData = JSON.stringify(row.Color).trim();
                    let colorsArrayRaw = colorsData.split(',');
                    for (let index = 0; index < colorsArrayRaw.length; index++) {
                        const element = this.cleanNameString(colorsArrayRaw[index]);
                        colorsArray.push(element);
                    }
                    dataObject.color = (colorsArray.length > 0) ? colorsArray : [];
                }
                const currentColor = await this.productColorVariantsRepository.findByFilterQuery({ productId: row.ProductId });
                if (colorsArray.length) {
                    const arrayCheck = await isArraySame(currentColor, colorsArray);
                    if (!arrayCheck) {
                        const currentVariants = (currentColor as ProductColorVariants[]).map((singleVariant: ProductColorVariants) =>
                            singleVariant.colorName
                        )
                        for (const color of colorsArray) {
                            if (!currentVariants.includes(color)) {
                                const colorData = await this.colorsRepository.findByColorName(color);
                                if (colorData) {
                                    if (Object.keys(colorData).length > 0) {
                                        const newColorEntity = {
                                            productId: row.ProductId,
                                            productName: row.Name,
                                            colorId: colorData[0]['id'],
                                            colorName: color,
                                            colorCode: colorData[0]['colorCode']
                                        }
                                        await this.productColorVariantsRepository.createEntity(newColorEntity)
                                    }
                                }
                            }
                        }
                    }
                }
                await this.updateProductStock(dataObject as UpdateProductDto, row.ProductId)
            }
        }
        return ResponseUtils.successResponseHandler(200, "Product stockin successfully completed!", "product_stockin_data", null);
    }

    async updateProductStock(dto: UpdateProductDto, productId: string): Promise<Product | Error> {
        const productEntity = {
            name: dto.name,
            price: dto.price,
            stock: dto.stock,
            weight: dto.weight,
            sku: dto.sku
        }
        const product = await this.repository.updateEntity(productId, productEntity);
        return product;
    }

    async updateProductRating(dto: UpdateProductDto, productId: string): Promise<Product | Error> {
        const productEntity = {
            averageRating: dto.averageRating,
            reviewCount: dto.reviewCount
        }
        const product = await this.repository.updateEntity(productId, productEntity);
        return product;
    }

    async importProductService(dto: CreateImportProductDto, file: Express.Multer.File): Promise<any> {
        let dumpID = uuidv4();
        const excelDataRow = await readXlsxFile(file.path).then((rows) => {
            let excelDataRowDump = [];
            rows.forEach(async element => {
                if (element[0] != "Name") {
                    let productRawData = {
                        'DumpID': dumpID,
                        'Name': element[0],
                        'Sku': element[1],
                        'Cost': element[2],
                        'Price': element[3],
                        'DiscountPrice': element[4],
                        'Stock': element[5],
                        'Weight': element[6],
                        'Color': element[7],
                        'ColorID': []
                    };
                    excelDataRowDump.push(productRawData);
                }
            });
            return excelDataRowDump;
        })

        for (let index = 0; index < excelDataRow.length; index++) {
            const rowsData = excelDataRow[index];

            // Colors Section
            let colorsArray = [];
            if (JSON.stringify(rowsData.Color).length > 0) {
                let colorsData = JSON.stringify(rowsData.Color).trim();
                colorsArray = colorsData.split(',')
            }

            for (let j = 0; j < colorsArray.length; j++) {
                const colorsName = colorsArray[j];
                let colorsRawData = await this.colorsRepository.findByFilterQuery({
                    colorName: this.cleanNameString(colorsName).trim()
                });
                if (!colorsRawData) {
                    let colorDto = {
                        colorName: this.cleanNameString(colorsName).trim(),
                        colorCode: '#0000'
                    };
                    let newColorData = await this.colorsRepository.createEntity(colorDto as CreateColorDto)
                    excelDataRow[index].ColorID.push(newColorData);
                } else {
                    if (Object.keys(colorsRawData).length == 0) {
                        let colorDto = {
                            colorName: this.cleanNameString(colorsName).trim(),
                            colorCode: '#0000'
                        };
                        let newColorData = await this.colorsRepository.createEntity(colorDto as CreateColorDto)
                        excelDataRow[index].ColorID.push(newColorData);
                    } else {
                        excelDataRow[index].ColorID.push(colorsRawData[0]);
                    }
                }
            }
        }

        excelDataRow.forEach(async row => {
            let dataRow = {
                'DumpID': row.DumpID,
                'Name': row.Name,
                'Sku': row.Sku,
                'Cost': row.Cost,
                'Price': row.Price,
                'DiscountPrice': row.DiscountPrice,
                'Stock': row.Stock,
                'Weight': row.Weight,
                'Color': row.Color
            };
            await this.productDumpImportRepository.createEntity(dataRow as CreateProductDumpImportDto);
        });

        const formattedResponse = {
            dumpID: dumpID,
            dump: excelDataRow
        }

        return ResponseUtils.successResponseHandler(200, "Import product dump data successfully!", "product_dump", formattedResponse);
    }

    async confirmImportProductService(dto: CreateConfirmImportProductDto): Promise<any> {
        if (dto.dumpID.length > 0) {
            let dumpData = await this.productDumpImportRepository.findByFilterQuery({
                DumpID: dto.dumpID
            });

            if (!dumpData) {
                throw new HttpException(
                    Constants.NOT_FOUND,
                    HttpStatus.NOT_FOUND
                );
            }

            for (let index = 0; index < Object.keys(dumpData).length; index++) {
                const row = dumpData[index];
                let productObject = {
                    'name': row.Name,
                    'sku': row.Sku,
                    'cost': row.Cost,
                    'price': row.Price,
                    'discountPrice': row.DiscountPrice,
                    'stock': row.Stock,
                    'weight': row.Weight,
                    'color': row.Color
                }

                // Colors Section
                let colorsArray = [];
                if (JSON.stringify(row.Color).length > 0) {
                    let colorsData = JSON.stringify(row.Color).trim();
                    let colorsArrayRaw = colorsData.split(',');

                    for (let index = 0; index < colorsArrayRaw.length; index++) {
                        const element = this.cleanNameString(colorsArrayRaw[index]);
                        colorsArray.push(element);
                    }
                    productObject.color = (colorsArray.length > 0) ? colorsArray : [];
                }
                await this.create(productObject as unknown as CreateProductDto, {})
            }
        }
        return ResponseUtils.successResponseHandler(200, "Product import successfully completed!", "product_dump_data", null);
    }

    async searchProductByName(name: string): Promise<Product | Error> {
        const data = await this.repository.searchProductByName(name);

        return ResponseUtils.successResponseHandler(200, "Product data fetched Successfully!", "data", data);
    }
}
